--***************************************************************
--* File         : IPLSPI.vhd                                   *
--* Created      : time-stamp: <08/03/30 17:43:59 kawasan>      *
--* Modified     : Time-stamp: <08/04/02 00:28:07 kawasan>      *
--*                                                             *
--* M25P80 -SPI-> IPLSPI -ParaBUS-> RAM                         *
--*                                                             *
--* Copyright 2008 Osamu Kawashima. all rights reserved	        *
--***************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity IPLSPI is
	generic (
		SADD_SIZE       : integer := 32;
		DADD_SIZE       : integer := 32
	);
	port (
		RESET           : in  std_logic;
		SYSCLK          : in  std_logic;

		S_SADD          : in  std_logic_vector(SADD_SIZE-1 downto 0);
		D_SADD          : in  std_logic_vector(DADD_SIZE-1 downto 0);
		D_EADD          : in  std_logic_vector(DADD_SIZE-1 downto 0);
		GO              : in  std_logic;
		DONE            : out std_logic;

		SPI_S_N         : out std_logic;	-- active=Low
		SPI_CLK         : out std_logic;
		SPI_DO          : out std_logic;
		SPI_DI          : in  std_logic;

		D_ADD           : out std_logic_vector(DADD_SIZE-1 downto 0);
		D_DO            : out std_logic_vector(7 downto 0);
		D_WE            : out std_logic
	);
end IPLSPI;

architecture RTL of IPLSPI is

	constant ZERO32		: std_logic_vector(31 downto 0) := (others=>'0');
	constant ONE32          : std_logic_vector(31 downto 0) := (others=>'1');

	type STATE is (S_STANDBY,S_NULL,
			S_RDID_START,S_RDID_INSTRUCTION,S_RDID_MANUFACTURER,S_RDID_STOP,
			S_FREAD_START,S_FREAD_INSTRUCTION,
			S_FREAD_ADDRESS0,S_FREAD_ADDRESS1,S_FREAD_ADDRESS2,
			S_FREAD_DUMMY,S_FREAD_DATAOUT,S_FREAD_MEMWR,S_FREAD_STOP);
	signal next_state       : STATE;
	signal return_state     : STATE;

	type SPI_STATE is (S_SPI_STANDBY,S_SPI_RUN);
	signal next_spi_state   : SPI_STATE;

	-- SPI INSTRUCTION
	constant C_SPI_INSTRUCTION_WREN  : std_logic_vector := x"06";
	constant C_SPI_INSTRUCTION_WRDI  : std_logic_vector := x"04";
	constant C_SPI_INSTRUCTION_RDID  : std_logic_vector := x"9F";
	constant C_SPI_INSTRUCTION_RDSR  : std_logic_vector := x"05";
	constant C_SPI_INSTRUCTION_WRSR  : std_logic_vector := x"01";
	constant C_SPI_INSTRUCTION_READ  : std_logic_vector := x"03";
	constant C_SPI_INSTRUCTION_FREAD : std_logic_vector := x"0B";
	constant C_SPI_INSTRUCTION_PP    : std_logic_vector := x"02";
	constant C_SPI_INSTRUCTION_SE    : std_logic_vector := x"D8";
	constant C_SPI_INSTRUCTION_BE    : std_logic_vector := x"C7";
	constant C_SPI_INSTRUCTION_DP    : std_logic_vector := x"B9";
	constant C_SPI_INSTRUCTION_RES   : std_logic_vector := x"AB";

	constant C_SPI_MANUFACTURER      : std_logic_vector := x"20";
	-- SPI INSTRUCTION

	signal spi_go           : std_logic;
	signal spi_done         : std_logic;
	signal spi_dat_tx       : std_logic_vector(7 downto 0);
	signal spi_dat_rx       : std_logic_vector(7 downto 0);
	signal spi_buf_tx       : std_logic_vector(7 downto 0);
	signal spi_buf_rx       : std_logic_vector(7 downto 0);
	signal spi_cnt          : std_logic_vector(2 downto 0);
	signal spi_out_en       : std_logic;
	signal spi_rd_en        : std_logic;

	signal d_add_is         : std_logic_vector(DADD_SIZE-1 downto 0);

begin

	-- State-Machine --
	process (SYSCLK)
	begin
		if SYSCLK'event and SYSCLK='1' then
			if RESET='1' then
				next_state <= S_STANDBY;
				return_state <= S_NULL;
				SPI_S_N <= '1';
				spi_dat_tx <= (others=>'0');
				spi_go <= '0';
				DONE <= '1';
				d_add_is <= (others=>'0');
				D_ADD <= (others=>'0');
				D_DO <= (others=>'0');
				D_WE <= '0';
			else
				case next_state is
					when S_STANDBY =>
						if GO='1' then
							next_state <= S_RDID_START;
							DONE <= '0';
							d_add_is <= D_SADD;
						else
							DONE <= '1';
						end if;

					when S_RDID_START =>
						next_state <= S_RDID_INSTRUCTION;
						SPI_S_N <= '0';
						spi_dat_tx <= C_SPI_INSTRUCTION_RDID;
						spi_go <= '1';

					when S_RDID_INSTRUCTION =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_RDID_MANUFACTURER;
							spi_dat_tx <= (others=>'1');
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_RDID_MANUFACTURER =>
						if spi_go='0' and spi_done='1' then
							if spi_dat_rx=C_SPI_MANUFACTURER then
								next_state <= S_RDID_STOP;
							else
								spi_go <= '1';
							end if;
						else
							spi_go <= '0';
						end if;

					when S_RDID_STOP =>
						next_state <= S_FREAD_START;
						SPI_S_N <= '1';

					when S_FREAD_START =>
						next_state <= S_FREAD_INSTRUCTION;
						SPI_S_N <= '0';
						spi_dat_tx <= C_SPI_INSTRUCTION_FREAD;
						spi_go <= '1';

					when S_FREAD_INSTRUCTION =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_ADDRESS0;
							spi_dat_tx <= S_SADD(23 downto 16);
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_ADDRESS0 =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_ADDRESS1;
							spi_dat_tx <= S_SADD(15 downto 8);
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_ADDRESS1 =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_ADDRESS2;
							spi_dat_tx <= S_SADD(7 downto 0);
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_ADDRESS2 =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_DUMMY;
							spi_dat_tx <= (others=>'1');
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_DUMMY =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_DATAOUT;
							spi_dat_tx <= (others=>'1');
							spi_go <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_DATAOUT =>
						if spi_go='0' and spi_done='1' then
							next_state <= S_FREAD_MEMWR;
							D_ADD <= d_add_is;
							D_DO <= spi_dat_rx;
							D_WE <= '1';
						else
							spi_go <= '0';
						end if;

					when S_FREAD_MEMWR =>
						if d_add_is=D_EADD then
							next_state <= S_FREAD_STOP;
						else
							next_state <= S_FREAD_DATAOUT;
							d_add_is <= d_add_is + 1;
							spi_dat_tx <= (others=>'1');
							spi_go <= '1';
						end if;
						D_WE <= '0';

					when S_FREAD_STOP =>
						next_state <= S_STANDBY;
						SPI_S_N <= '1';
						DONE <= '1';

					when others  =>
						next_state <= S_STANDBY;
				end case;
			end if;
		end if;
	end process;
	-- State-Machine --

	-- State-Machine(SPI-OUT) --
	process (SYSCLK)
	begin
		if SYSCLK'event and SYSCLK='1' then
			if RESET='1' then
				SPI_DO <= '1';
				spi_done <= '1';
				spi_cnt <= "111";
				spi_buf_tx <= (others=>'0');
				spi_rd_en <= '0';
				spi_out_en <= '0';
			else
				case next_spi_state is
					when S_SPI_STANDBY =>
						if spi_go='1' then
							next_spi_state <= S_SPI_RUN;
							SPI_DO <= spi_dat_tx(7);
							spi_done <= '0';
							spi_cnt <= "110";
							spi_buf_tx <= spi_dat_tx(6 downto 0) & '0';
							spi_rd_en <= '1';
							spi_out_en <= '1';
						else
							SPI_DO <= '1';
							spi_done <= '1';
							spi_cnt <= "111";
							spi_buf_tx <= (others=>'0');
							spi_rd_en <= '0';
							spi_out_en <= '0';
						end if;

					when S_SPI_RUN =>
						if spi_cnt=0 then
							next_spi_state <= S_SPI_STANDBY;
							spi_done <= '1';
						else
							spi_cnt <= spi_cnt - 1;
						end if;
						SPI_DO <= spi_buf_tx(7);
						spi_buf_tx <= spi_buf_tx(6 downto 0) & '0';

					when others  =>
						next_spi_state <= S_SPI_STANDBY;
				end case;
			end if;
		end if;
	end process;
	-- State-Machine(SPI-OUT) --

	process (SYSCLK)
	begin
		if SYSCLK'event and SYSCLK='0' then
			if RESET='1' or spi_rd_en='0' then
				spi_buf_rx <= (others=>'0');
			else
				spi_buf_rx <= spi_buf_rx(6 downto 0) & SPI_DI;
			end if;
		end if;
	end process;

	spi_dat_rx <= spi_buf_rx;

	SPI_CLK <= not SYSCLK when spi_out_en='1' else '1';

end RTL;
